import abc
from re import L
from typing import Optional, List, Tuple, Dict, Callable
from unicodedata import name

import jax.numpy as jnp
from chex import dataclass
from jax import vmap

from gpjax.types import Array


@dataclass(repr=False)
class Kernel:
    active_dims: Optional[List[int]] = None
    stationary: Optional[bool] = False
    spectral: Optional[bool] = False
    name: Optional[str] = "Kernel"
    _params: Optional[Dict] = None

    def __post_init__(self):
        self.ndims = 1 if not self.active_dims else len(self.active_dims)

    @abc.abstractmethod
    def __call__(self, x: Array, y: Array, params: dict) -> Array:
        raise NotImplementedError

    def slice_input(self, x: Array) -> Array:
        return x[..., self.active_dims]

    @property
    def ard(self):
        return True if self.ndims > 1 else False

    @property
    def params(self) -> dict:
        return self._params

    @params.setter
    def params(self, value):
        self._params = value


@dataclass
class CombinationKernel:
    kernel_set: List[Kernel]
    name: Optional[str] = "Combination kernel"
    combination_fn: Optional[Callable] = None

    @property
    def params(self) -> List[Dict]:
        return [kernel.params for kernel in self.kernel_set]

    def __call__(self, x: Array, y: Array, params: dict) -> Array:
        return self.combination_fn(
            jnp.stack([k(x, y, p) for k, p in zip(self.kernel_set, params)])
        )


@dataclass
class SumKernel(CombinationKernel):
    name: Optional[str] = "Sum kernel"
    combination_fn: Optional[Callable] = jnp.sum


@dataclass
class ProductKernel(CombinationKernel):
    name: Optional[str] = "Product kernel"
    combination_fn: Optional[Callable] = jnp.prod


@dataclass(repr=False)
class RBF(Kernel):
    name: Optional[str] = "Radial basis function kernel"

    def __post_init__(self):
        self.ndims = 1 if not self.active_dims else len(self.active_dims)
        self._params = {
            "lengthscale": jnp.array([1.0] * self.ndims),
            "variance": jnp.array([1.0]),
        }

    def __call__(
        self, x: jnp.DeviceArray, y: jnp.DeviceArray, params: dict
    ) -> Array:
        x = self.slice_input(x) / params["lengthscale"]
        y = self.slice_input(y) / params["lengthscale"]
        K = params["variance"] * jnp.exp(-0.5 * squared_distance(x, y))
        return K.squeeze()


@dataclass(repr=False)
class Matern12(Kernel):
    name: Optional[str] = "Matern 1/2"

    def __post_init__(self):
        self.ndims = 1 if not self.active_dims else len(self.active_dims)
        self._params = {
            "lengthscale": jnp.array([1.0] * self.ndims),
            "variance": jnp.array([1.0]),
        }

    def __call__(
        self, x: jnp.DeviceArray, y: jnp.DeviceArray, params: dict
    ) -> Array:
        x = self.slice_input(x) / params["lengthscale"]
        y = self.slice_input(y) / params["lengthscale"]
        K = params["variance"] * jnp.exp(-0.5 * euclidean_distance(x, y))
        return K.squeeze()


@dataclass(repr=False)
class Matern32(Kernel):
    name: Optional[str] = "Matern 3/2"

    def __post_init__(self):
        self.ndims = 1 if not self.active_dims else len(self.active_dims)
        self._params = {
            "lengthscale": jnp.array([1.0] * self.ndims),
            "variance": jnp.array([1.0]),
        }

    def __call__(
        self, x: jnp.DeviceArray, y: jnp.DeviceArray, params: dict
    ) -> Array:
        x = self.slice_input(x) / params["lengthscale"]
        y = self.slice_input(y) / params["lengthscale"]
        tau = euclidean_distance(x, y)
        K = (
            params["variance"]
            * (1.0 + jnp.sqrt(3.0) * tau)
            * jnp.exp(-jnp.sqrt(3.0) * tau)
        )
        return K.squeeze()


@dataclass(repr=False)
class Matern52(Kernel):
    name: Optional[str] = "Matern 5/2"

    def __post_init__(self):
        self.ndims = 1 if not self.active_dims else len(self.active_dims)
        self._params = {
            "lengthscale": jnp.array([1.0] * self.ndims),
            "variance": jnp.array([1.0]),
        }

    def __call__(
        self, x: jnp.DeviceArray, y: jnp.DeviceArray, params: dict
    ) -> Array:
        x = self.slice_input(x) / params["lengthscale"]
        y = self.slice_input(y) / params["lengthscale"]
        tau = euclidean_distance(x, y)
        K = (
            params["variance"]
            * (1.0 + jnp.sqrt(5.0) * tau + 5.0 / 3.0 * jnp.square(tau))
            * jnp.exp(-jnp.sqrt(5.0) * tau)
        )
        return K.squeeze()


@dataclass(repr=False)
class Polynomial(Kernel):
    name: Optional[str] = "Polynomial"
    degree: int = 1

    def __post_init__(self):
        self.ndims = 1 if not self.active_dims else len(self.active_dims)
        self._params = {
            "shift": jnp.array([1.0]),
            "variance": jnp.array([1.0] * self.ndims),
        }
        self.name = f"Polynomial Degree: {self.degree}"

    def __call__(
        self, x: jnp.DeviceArray, y: jnp.DeviceArray, params: dict
    ) -> Array:
        x = self.slice_input(x).squeeze()
        y = self.slice_input(y).squeeze()
        K = jnp.power(
            params["shift"] + jnp.dot(x * params["variance"], y), self.degree
        )
        return K.squeeze()


def squared_distance(x: Array, y: Array):
    return jnp.sum((x - y) ** 2)


def euclidean_distance(x: Array, y: Array):
    return jnp.sum(jnp.abs(x - y))


def gram(kernel: Kernel, inputs: Array, params: dict) -> Array:
    return vmap(lambda x1: vmap(lambda y1: kernel(x1, y1, params))(inputs))(
        inputs
    )


def cross_covariance(kernel: Kernel, x: Array, y: Array, params: dict) -> Array:
    return vmap(lambda x1: vmap(lambda y1: kernel(x1, y1, params))(x))(y)



import typing as tp

import jax
import jax.numpy as jnp
import optax
from tqdm import trange


def fit(
    objective,
    params: dict,
    opt_init,
    opt_update,
    get_params,
    n_iters: int = 100,
    log_rate: int = 10,
) -> tp.Dict:
    opt_state = opt_init(params)

    def step(i, opt_state):
        params = get_params(opt_state)
        v, g = jax.value_and_grad(objective)(params)
        return opt_update(i, g, opt_state), v

    tr = trange(n_iters)
    for i in tr:
        opt_state, val = step(i, opt_state)
        if i % log_rate == 0 or i == n_iters:
            tr.set_postfix({"Objective": jnp.round(val, 2)})
    return get_params(opt_state)


def optax_fit(
    objective,
    params: dict,
    optax_optim,
    n_iters: int = 100,
    log_rate: int = 10,
) -> tp.Dict:
    opt_state = optax_optim.init(params)
    ps = params

    def step(params, opt_state):
        v, g = jax.value_and_grad(objective)(params)
        updates, opt_state = optax_optim.update(g, opt_state, params)
        params = optax.apply_updates(params, updates)
        return params, opt_state, v

    tr = trange(n_iters)
    for i in tr:
        params, opt_state, val = step(params, opt_state)
        if i % log_rate == 0 or i == n_iters:
            tr.set_postfix({"Objective": jnp.round(val, 2)})
    return params



import typing as tp

import jax.numpy as jnp
import numpy as np
from chex import dataclass

NoneType = type(None)
Array = tp.Union[np.ndarray, jnp.ndarray]


@dataclass(repr=False)
class Dataset:
    X: Array
    y: Array = None

    def __repr__(self) -> str:
        return (
            f"- Number of datapoints: {self.X.shape[0]}\n- Dimension:"
            f" {self.X.shape[1]}"
        )

    @property
    def n(self) -> int:
        return self.X.shape[0]

    @property
    def in_dim(self) -> int:
        return self.X.shape[1]

    @property
    def out_dim(self) -> int:
        return self.y.shape[1]


def verify_dataset(ds: Dataset) -> None:
    assert ds.X.ndim == 2, (
        "2-dimensional training inputs are required. Current dimension:"
        f" {ds.X.ndim}."
    )
    if ds.y is not None:
        assert ds.y.ndim == 2, (
            "2-dimensional training outputs are required. Current dimension:"
            f" {ds.y.ndim}."
        )
        assert ds.X.shape[0] == ds.y.shape[0], (
            "Number of inputs must equal the number of outputs. \nCurrent"
            f" counts:\n- X: {ds.X.shape[0]}\n- y: {ds.y.shape[0]}"
        )



import typing as tp
from abc import abstractmethod, abstractproperty

import jax.numpy as jnp
import tensorflow_probability.substrates.jax.distributions as tfd
from chex import dataclass
from jax.scipy.linalg import cho_factor, cho_solve, solve_triangular

from gpjax.config import get_defaults

from .kernels import Kernel, cross_covariance, gram
from .likelihoods import (
    Gaussian,
    Likelihood,
    NonConjugateLikelihoods,
    NonConjugateLikelihoodType,
)
from .mean_functions import MeanFunction, Zero
from .parameters import evaluate_priors, copy_dict_structure, transform
from .types import Array, Dataset
from .utils import I, concat_dictionaries


@dataclass
class GP:
    @abstractmethod
    def mean(self) -> tp.Callable[[Dataset], Array]:
        raise NotImplementedError

    @abstractmethod
    def variance(self) -> tp.Callable[[Dataset], Array]:
        raise NotImplementedError

    @abstractproperty
    def params(self) -> tp.Dict:
        raise NotImplementedError


#######################
# GP Priors
#######################
@dataclass(repr=False)
class Prior(GP):
    kernel: Kernel
    mean_function: tp.Optional[MeanFunction] = Zero()
    name: tp.Optional[str] = "Prior"

    def __mul__(self, other: Gaussian):
        return construct_posterior(prior=self, likelihood=other)

    def mean(self, params: dict) -> tp.Callable[[Array], Array]:
        def mean_fn(test_points: Array):
            mu = self.mean_function(test_points, params["mean_function"])
            return mu

        return mean_fn

    def variance(self, params: dict) -> tp.Callable[[Array], Array]:
        def variance_fn(test_points: Array):
            Kff = gram(self.kernel, test_points, params["kernel"])
            jitter_matrix = I(test_points.shape[0]) * 1e-8
            covariance_matrix = Kff + jitter_matrix
            return covariance_matrix

        return variance_fn

    @property
    def params(self) -> dict:
        return {
            "kernel": self.kernel.params,
            "mean_function": self.mean_function.params,
        }

    def random_variable(
        self, test_points: Array, params: dict
    ) -> tfd.Distribution:
        n = test_points.shape[0]
        mu = self.mean(params)(test_points)
        sigma = self.variance(params)(test_points)
        sigma += I(n) * 1e-8
        return tfd.MultivariateNormalTriL(
            mu.squeeze(), jnp.linalg.cholesky(sigma)
        )


#######################
# GP Posteriors
#######################
@dataclass
class Posterior(GP):
    prior: Prior
    likelihood: Likelihood
    name: tp.Optional[str] = "GP Posterior"

    @abstractmethod
    def mean(
        self, training_data: Dataset, params: dict
    ) -> tp.Callable[[Dataset], Array]:
        raise NotImplementedError

    @abstractmethod
    def variance(
        self, training_data: Dataset, params: dict
    ) -> tp.Callable[[Dataset], Array]:
        raise NotImplementedError

    @property
    def params(self) -> dict:
        return concat_dictionaries(
            self.prior.params, {"likelihood": self.likelihood.params}
        )


@dataclass
class ConjugatePosterior(Posterior):
    prior: Prior
    likelihood: Gaussian
    name: tp.Optional[str] = "ConjugatePosterior"

    def mean(
        self, training_data: Dataset, params: dict
    ) -> tp.Callable[[Array], Array]:
        X, y = training_data.X, training_data.y
        sigma = params["likelihood"]["obs_noise"]
        n_train = training_data.n
        # Precompute covariance matrices
        Kff = gram(self.prior.kernel, X, params["kernel"])
        prior_mean = self.prior.mean_function(X, params["mean_function"])
        L = cho_factor(Kff + I(n_train) * sigma, lower=True)

        prior_distance = y - prior_mean
        weights = cho_solve(L, prior_distance)

        def mean_fn(test_inputs: Array) -> Array:
            prior_mean_at_test_inputs = self.prior.mean_function(
                test_inputs, params["mean_function"]
            )
            Kfx = cross_covariance(
                self.prior.kernel, X, test_inputs, params["kernel"]
            )
            return prior_mean_at_test_inputs + jnp.dot(Kfx, weights)

        return mean_fn

    def variance(
        self, training_data: Dataset, params: dict
    ) -> tp.Callable[[Array], Array]:
        X = training_data.X
        n_train = training_data.n
        obs_noise = params["likelihood"]["obs_noise"]
        n_train = training_data.n
        Kff = gram(self.prior.kernel, X, params["kernel"])
        Kff += I(n_train) * 1e-8
        L = cho_factor(Kff + I(n_train) * obs_noise, lower=True)

        def variance_fn(test_inputs: Array) -> Array:
            Kfx = cross_covariance(
                self.prior.kernel, X, test_inputs, params["kernel"]
            )
            Kxx = gram(self.prior.kernel, test_inputs, params["kernel"])
            latent_values = cho_solve(L, Kfx.T)
            return Kxx - jnp.dot(Kfx, latent_values)

        return variance_fn

    def marginal_log_likelihood(
        self,
        training: Dataset,
        transformations: tp.Dict,
        priors: dict = None,
        static_params: dict = None,
        negative: bool = False,
    ) -> tp.Callable[[Dataset], Array]:
        x, y = training.X, training.y

        def mll(
            params: dict,
        ):
            params = transform(params=params, transform_map=transformations)
            if static_params:
                params = concat_dictionaries(params, transform(static_params))
            mu = self.prior.mean_function(x, params)
            gram_matrix = gram(self.prior.kernel, x, params["kernel"])
            gram_matrix += params["likelihood"]["obs_noise"] * I(x.shape[0])
            L = jnp.linalg.cholesky(gram_matrix)
            random_variable = tfd.MultivariateNormalTriL(mu, L)

            log_prior_density = evaluate_priors(params, priors)
            constant = jnp.array(-1.0) if negative else jnp.array(1.0)
            return constant * (
                random_variable.log_prob(y.squeeze()).mean() + log_prior_density
            )

        return mll


@dataclass
class NonConjugatePosterior(Posterior):
    prior: Prior
    likelihood: NonConjugateLikelihoodType
    name: tp.Optional[str] = "Non-Conjugate Posterior"

    def __repr__(self):
        mean_fn_string = self.prior.mean_function.__repr__()
        kernel_string = self.prior.kernel.__repr__()
        likelihood_string = self.likelihood.__repr__()
        return (
            f"Conjugate Posterior\n{'-'*80}\n- {mean_fn_string}\n-"
            f" {kernel_string}\n- {likelihood_string}"
        )

    @property
    def params(self) -> dict:
        hyperparameters = concat_dictionaries(
            self.prior.params, {"likelihood": self.likelihood.params}
        )
        hyperparameters["latent"] = jnp.zeros(
            shape=(self.likelihood.num_datapoints, 1)
        )
        return hyperparameters

    def mean(
        self, training_data: Dataset, params: dict
    ) -> tp.Callable[[Dataset], Array]:
        X, y = training_data.X, training_data.y
        N = training_data.n
        Kff = gram(self.prior.kernel, X, params["kernel"])
        L = jnp.linalg.cholesky(Kff + I(N) * 1e-6)

        def meanf(test_inputs: Array) -> Array:
            Kfx = cross_covariance(
                self.prior.kernel, X, test_inputs, params["kernel"]
            )
            Kxx = gram(self.prior.kernel, test_inputs, params["kernel"])
            A = solve_triangular(L, Kfx.T, lower=True)
            latent_var = Kxx - jnp.sum(jnp.square(A), -2)
            latent_mean = jnp.matmul(A.T, params["latent"])

            lvar = jnp.diag(latent_var)

            moment_fn = self.likelihood.predictive_moment_fn
            pred_rv = moment_fn(latent_mean.ravel(), lvar)
            return pred_rv.mean().reshape(-1, 1)

        return meanf

    def variance(
        self, training_data: Dataset, params: dict
    ) -> tp.Callable[[Dataset], Array]:
        X, y = training_data.X, training_data.y
        N = training_data.n
        Kff = gram(self.prior.kernel, X, params["kernel"])
        L = jnp.linalg.cholesky(Kff + I(N) * 1e-6)

        def variancef(test_inputs: Array) -> Array:
            Kfx = cross_covariance(
                self.prior.kernel, X, test_inputs, params["kernel"]
            )
            Kxx = gram(self.prior.kernel, test_inputs, params["kernel"])
            A = solve_triangular(L, Kfx.T, lower=True)
            latent_var = Kxx - jnp.sum(jnp.square(A), -2)
            latent_mean = jnp.matmul(A.T, params["latent"])

            lvar = jnp.diag(latent_var)

            moment_fn = self.likelihood.predictive_moment_fn
            pred_rv = moment_fn(latent_mean.ravel(), lvar)
            return jnp.diag(pred_rv.variance())

        return variancef

    def marginal_log_likelihood(
        self,
        training: Dataset,
        transformations: tp.Dict,
        priors: dict = None,
        static_params: dict = None,
        negative: bool = False,
    ) -> tp.Callable[[Dataset], Array]:
        x, y = training.X, training.y
        n = training.n
        jitter = get_defaults()["jitter"]
        if not priors:
            priors = copy_dict_structure(self.params)
            priors["latent"] = tfd.Normal(loc=0.0, scale=1.0)

        def mll(params: dict):
            params = transform(params=params, transform_map=transformations)
            if static_params:
                params = concat_dictionaries(params, transform(static_params))
            gram_matrix = gram(self.prior.kernel, x, params["kernel"])
            gram_matrix += I(n) * jitter
            L = jnp.linalg.cholesky(gram_matrix)
            F = jnp.matmul(L, params["latent"])
            rv = self.likelihood.link_function(F)
            ll = jnp.sum(rv.log_prob(y))

            log_prior_density = evaluate_priors(params, priors)
            constant = jnp.array(-1.0) if negative else jnp.array(1.0)
            return constant * (ll + log_prior_density)

        return mll


def construct_posterior(prior: Prior, likelihood: Likelihood) -> Posterior:
    if isinstance(likelihood, Gaussian):
        PosteriorGP = ConjugatePosterior
    elif any([isinstance(likelihood, l) for l in NonConjugateLikelihoods]):
        PosteriorGP = NonConjugatePosterior
    else:
        raise NotImplementedError(
            f"No posterior implemented for {likelihood.name} likelihood"
        )
    return PosteriorGP(prior=prior, likelihood=likelihood)



import abc
from typing import Callable, Optional

import jax.numpy as jnp
from chex import dataclass
from tensorflow_probability.substrates.jax import distributions as tfd


@dataclass(repr=False)
class Likelihood:
    num_datapoints: int  # The number of datapoints that the likelihood factorises over
    name: Optional[str] = "Likelihood"

    def __repr__(self):
        return f"{self.name} likelihood function"

    @property
    @abc.abstractmethod
    def params(self) -> dict:
        raise NotImplementedError

    @property
    @abc.abstractmethod
    def link_function(self) -> Callable:
        raise NotImplementedError


@dataclass(repr=False)
class Gaussian(Likelihood):
    name: Optional[str] = "Gaussian"

    @property
    def params(self) -> dict:
        return {"obs_noise": jnp.array([1.0])}

    @property
    def link_function(self) -> Callable:
        def identity_fn(x):
            return x

        return identity_fn


@dataclass(repr=False)
class Bernoulli(Likelihood):
    name: Optional[str] = "Bernoulli"

    @property
    def params(self) -> dict:
        return {}

    @property
    def link_function(self) -> Callable:
        def link_fn(x):
            return tfd.ProbitBernoulli(x)

        return link_fn

    @property
    def predictive_moment_fn(self) -> Callable:
        def moment_fn(mean: jnp.DeviceArray, variance: jnp.DeviceArray):
            rv = self.link_function(mean / jnp.sqrt(1 + variance))
            return rv

        return moment_fn


NonConjugateLikelihoods = [Bernoulli]
NonConjugateLikelihoodType = Bernoulli  # Union[Bernoulli]



import typing as tp
import warnings
from copy import deepcopy

import jax
import jax.numpy as jnp
import tensorflow_probability.substrates.jax as tfp
from tensorflow_probability.substrates.jax import distributions as tfd

from .config import get_defaults
from .types import Array


################################
# Base operations
################################
def initialise(obj) -> tp.Tuple[tp.Dict, tp.Dict, tp.Dict]:
    params = obj.params
    constrainers, unconstrainers = build_transforms(params)
    return params, constrainers, unconstrainers


def recursive_items(d1, d2):
    for key, value in d1.items():
        if type(value) is dict:
            yield from recursive_items(value, d2[key])
        else:
            yield (key, value, d2[key])


def recursive_complete(d1, d2) -> dict:
    for key, value in d1.items():
        if type(value) is dict:
            if key in d2.keys():
                recursive_complete(value, d2[key])
            # else:
            #     pass
        else:
            if key in d2.keys():
                d1[key] = d2[key]
    return d1


# def recursive_fn(d1, d2, fn: tp.Callable[[tp.Any], tp.Any]):
#     for key, value in d1.items():
#         if type(value) is dict:
#             yield from recursive_fn(value, d2[key], fn)
#         else:
#             yield fn(value, d2[key])


################################
# Parameter transformation
################################
def build_transforms(params, key=None) -> tp.Tuple[tp.Dict, tp.Dict]:
    constrainer, unconstrainer = copy_dict_structure(params), copy_dict_structure(params)
    config = get_defaults()
    transform_set = config["transformations"]

    def recursive_transforms(ps, cs, ucs) -> tp.Tuple[tp.Dict, tp.Dict]:
        for key, value in ps.items():
            if type(value) is dict:
                recursive_transforms(value, cs[key], ucs[key])
            else:
                if key in transform_set.keys():
                    transform_type = transform_set[key]
                    bijector = transform_set[transform_type]
                else:
                    bijector = tfp.bijectors.Identity()
                    warnings.warn(
                        f"Parameter {key} has no transform. Defaulting to"
                        " identity transfom."
                    )
                cs[key] = bijector.forward
                ucs[key] = bijector.inverse
        return cs, ucs

    constrainers, unconstrainers = recursive_transforms(
        params, constrainer, unconstrainer
    )
    return constrainers, unconstrainers


def transform(params: dict, transform_map: dict):
    transformed_params = copy_dict_structure(params)

    def apply_transform(untransformed_params, transformed_params, transform_map):
        for key, value in untransformed_params.items():
            if type(value) is dict:
                apply_transform(value, transformed_params[key], transform_map[key])
            else:
                transformed_params[key] = transform_map[key](value)
        return transformed_params

    return apply_transform(params, transformed_params, transform_map)


################################
# Priors
################################
def log_density(param: jnp.DeviceArray, density: tfd.Distribution) -> Array:
    if type(density) == type(None):
        log_prob = jnp.array(0.0)
    else:
        log_prob = jnp.sum(density.log_prob(param))
    return log_prob


def copy_dict_structure(params: dict) -> dict:
    # Copy dictionary structure
    prior_container = deepcopy(params)
    # Set all values to zero
    prior_container = jax.tree_map(lambda _: None, prior_container)
    return prior_container


def structure_priors(params: dict, priors: dict) -> dict:
    """First create a dictionary with equal structure to the parameters. Then, for each supplied prior, overwrite the None value if it exists.

    Args:
        params (dict): [description]
        priors (dict): [description]

    Returns:
        dict: [description]
    """
    prior_container = copy_dict_structure(params)
    # Where a prior has been supplied, override the None value by the prior distribution.
    complete_prior = recursive_complete(prior_container, priors)
    return complete_prior


def evaluate_priors(params: dict, priors: dict) -> dict:
    """Recursive loop over pair of dictionaries that correspond to a parameter's
    current value and the parameter's respective prior distribution. For
    parameters where a prior distribution is specified, the log-prior density is
    evaluated at the parameter's current value.

    Args: params (dict): Dictionary containing the current set of parameter
        estimates. priors (dict): Dictionary specifying the parameters' prior
        distributions.

    Returns: Array: The log-prior density, summed over all parameters.
    """
    lpd = jnp.array(0.0)
    if priors is not None:
        for name, param, prior in recursive_items(params, priors):
            lpd += log_density(param, prior)
    return lpd


def prior_checks(priors: dict) -> dict:
    if "latent" in priors.keys():
        latent_prior = priors["latent"]
        if isinstance(latent_prior, tfd.Distribution) and latent_prior.name != "Normal":
            warnings.warn(
                f"A {latent_prior.name} distribution prior has been placed on"
                " the latent function. It is strongly advised that a"
                " unit-Gaussian prior is used."
            )
        else:
            if not latent_prior:
                priors["latent"] = tfd.Normal(loc=0.0, scale=1.0)
    else:
        priors["latent"] = tfd.Normal(loc=0.0, scale=1.0)

    return priors



import abc
from typing import Dict, Optional

import jax.numpy as jnp
from chex import dataclass

from .types import Array


@dataclass(repr=False)
class MeanFunction:
    output_dim: Optional[int] = 1
    name: Optional[str] = "Mean function"

    @abc.abstractmethod
    def __call__(self, x: Array) -> Array:
        raise NotImplementedError

    def __repr__(self):
        return f"{self.name}\n\t Output dimension: {self.output_dim}"

    @property
    @abc.abstractmethod
    def params(self) -> dict:
        raise NotImplementedError


@dataclass(repr=False)
class Zero(MeanFunction):
    output_dim: Optional[int] = 1
    name: Optional[str] = "Zero mean function"

    def __call__(self, x: Array, params: dict) -> Array:
        out_shape = (x.shape[0], self.output_dim)
        return jnp.zeros(shape=out_shape)

    @property
    def params(self) -> dict:
        return {}


@dataclass(repr=False)
class Constant(MeanFunction):
    output_dim: Optional[int] = 1
    name: Optional[str] = "Constant mean function"

    def __call__(self, x: Array, params: Dict) -> Array:
        out_shape = (x.shape[0], self.output_dim)
        return jnp.ones(shape=out_shape) * params["variance"]

    @property
    def params(self) -> dict:
        return {"variance": jnp.array(1.0)}



from typing import Tuple

import jax.random as jr
import tensorflow_probability.substrates.jax.bijectors as tfb
from ml_collections import ConfigDict


__config = None


def get_defaults() -> ConfigDict:
    config = ConfigDict()
    config.key = jr.PRNGKey(123)
    # Covariance matrix stabilising jitter
    config.jitter = 1e-6

    # Default bijections
    config.transformations = transformations = ConfigDict()
    transformations.positive_transform = tfb.Softplus()
    transformations.identity_transform = tfb.Identity()

    # Default parameter transforms
    transformations.lengthscale = "positive_transform"
    transformations.shift = "positive_transform"
    transformations.variance = "positive_transform"
    transformations.obs_noise = "positive_transform"
    transformations.latent = "identity_transform"
    transformations.basis_fns = "identity_transform"
    transformations.offset = "identity_transform"
    global __config
    if not __config:
        __config = config
    return __config


def add_parameter(param_name: str, bijection: tfb.Bijector) -> None:
    lookup_name = f"{param_name}_transform"
    get_defaults()
    __config.transformations[lookup_name] = bijection
    __config.transformations[param_name] = lookup_name
